/*
 * include/haproxy/quic_trace-t.h
 * Definitions for QUIC CLI internal types, constants and flags.
 *
 * Copyright (C) 2023
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */
#ifndef _HAPROXY_QUIC_CLI_T_H
#define _HAPROXY_QUIC_CLI_T_H

extern unsigned int qc_epoch;

#endif /* _HAPROXY_QUIC_CLI_T_H */
